from astropy.nddata import support_nddata, NDDataRef
from astropy.table import Table
from astropy import log
import numpy as np
import astropy.units as u

from smoothy.core import fix_mask, integrate
from smoothy.upi.axes import spectral_velocities


@support_nddata
def _moment(data,order,wcs=None,mask=None,unit=None,restfrq=None):
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    data = fix_mask(data,mask)
    dim = wcs.wcs.spec
    rdim = data.ndim - 1 - dim
    v = spectral_velocities(data, wcs, fqis=np.arange(data.shape[rdim]), restfrq=restfrq)
    v = v.value
    m0 = data.sum(axis=rdim)
    if order == 0:
        mywcs = wcs.dropaxis(dim)
        return NDDataRef(m0.data, uncertainty=None, mask=m0.mask,wcs=mywcs, meta=None, unit=unit)
    #mu,alpha=np.average(data,axis=rdim,weights=v,returned=True)
    mu,alpha = np.ma.average(data, axis=rdim, weights=v, returned=True)
    m1=alpha*mu/m0
    if order == 1:
        mywcs = wcs.dropaxis(dim)
        return NDDataRef(m1.data, uncertainty=None, mask=m1.mask,wcs=mywcs, meta=None, unit=u.km/u.s)
    v2 = v*v
    var, beta = np.ma.average(data, axis=rdim, weights=v2, returned=True)
    #var,beta=data.average(axis=rdim,weights=v2,returned=True)
    m2 = np.sqrt(beta*var/m0 - m1*m1)
    if order == 2:
        mywcs = wcs.dropaxis(dim)
        return NDDataRef(m2.data, uncertainty=None, mask=m2.mask, wcs=mywcs, meta=None, unit=u.km*u.km/u.s/u.s)
    log.error("Order not supported")
    return None


# Should return a NDData
@support_nddata
def moment0(data, wcs=None, mask=None, unit=None, restfrq=None):
    """
        Calculate moment 0 from a data cube.

        Parameters
        ----------
        data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.
        mask : numpy.ndarray
            Mask for data.
        unit : astropy.units.Unit
            Astropy unit (http://docs.astropy.org/en/stable/units/).
        restfrq : astropy.units.quantity.Quantity
            Rest frequency

        Returns
        -------
        result: astropy.nddata.NDDataRef
            Moment 0 of the data cube

    """
    return _moment(data, 0, wcs, mask, unit, restfrq)


@support_nddata
def moment1(data, wcs=None, mask=None, unit=None, restfrq=None):
    """
        Calculate moment 1 from a data cube.

        Parameters
        ----------
        data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use
        mask : numpy.ndarray
            Mask for data.
        unit : astropy.units.Unit
            Astropy unit (http://docs.astropy.org/en/stable/units/)
        restfrq : astropy.units.quantity.Quantity
            Rest frequency

        Returns
        -------
        result: astropy.nddata.NDData
            Moment 1 of the data cube

    """
    return _moment(data, 1, wcs, mask, unit, restfrq)


@support_nddata
def moment2(data, wcs=None, mask=None, unit=None, restfrq=None):
    """
        Calculate moment 2 from a data cube.

        Parameters
        ----------
        data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use
        mask : numpy.ndarray
            Mask for data.
        unit : astropy.units.Unit
            Astropy unit (http://docs.astropy.org/en/stable/units/)
        restfrq : astropy.units.quantity.Quantity
            Rest frequency

        Returns
        -------
        result: astropy.nddata.NDDataRef
            Moment 2 of the data cube

    """
    return _moment(data, 2, wcs, mask, unit, restfrq)


# TODO: Fix this function, is not working correctly
@support_nddata
def spectra(data, wcs=None, mask=None, unit=None, restrict=None):
    """


        Parameters
        ----------
        data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use
        mask : numpy.ndarray
            Mask for data.
        unit : astropy.units.Unit
            Astropy unit (http://docs.astropy.org/en/stable/units/)
        restrict : boolean


        Returns
        -------
        result: astropy.nddata.NDData
            Moment 2 of the data cube

    """
    if restrict is None:
        #Create NDD and WCS change...
        return integrate(data,axis=(1,2))
    else:
        log.error("Not Implemented Yet!")

        # Get 1 pixel aperture
        aperture=np.abs(wcs.celestial.wcs.cdelt[0])*u.deg
    #if position.unit == u.pix and aperture.unit == u.pix:
    #    # TODO:  Here is the nasty part
    #    lb=np.array([0,            position[1].value - aperture.value, position[0].value - aperture.value])
    #    ub=np.array([data.shape[2],position[1].value + aperture.value, position[0].value + aperture.value])
    #else:
    #    log.error("Not Implemented Yet!")
    #specview=data[slab(data,lb,ub)]
    #return specview.sum(axis=(1,2))



from astropy.table import Table, Column
import numpy as np
from .axes import axes_names


def _pix_table_creator(values,wcs):
    tab = Table()
    names = axes_names(None, wcs)
    for i in range(names.size):
       tab[names[i]]=Column(values[:,i],unit=u.pix)
    return tab


def _world_table_creator(values,wcs):
    uvec=np.array(wcs.wcs.cunit)[::-1]
    values=np.fliplr(values)
    tab = Table()
    names = axes_names(None, wcs)
    for i in range(names.size):
       tab[names[i]]=Column(values[:,i],unit=uvec[i])
    return tab


def _unitize(vec,wcs):
    uvec=np.array(wcs.wcs.cunit)[::-1]
    return vec*uvec



import numpy as np
import astropy.units as u
from astropy.nddata import support_nddata
from astropy import log

from smoothy import core

# axes_names used in formatting
# from smoothy.upi.data import Data


@support_nddata
def axes_names(data, wcs=None):
    """
        Get the axes's names.

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.

        Returns
        -------
        result: numpy.ndarray
            Numpy ndarray with the axes's names from the WCS.

    """
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    return np.array(wcs.axis_type_names)[::-1]


from smoothy.upi.formatting import _unitize, _world_table_creator


@support_nddata
def extent(data, wcs=None, region=None):
    """
        Get the axes extent.

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.
        region :(lower : (M,N) or (M,N,Z), upper : (M,N) or (M,N,Z))
            Start and End index in data (int tuples)

        Returns
        -------
        result: (M, N) tuple of astropy.units.quantity.Quantity
            Axes extent

    """
    # TODO: These can be a decorator
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None

    if region is None:
        lower=None
        upper=None
    else:
        (lower,upper) = region

    if lower == None:
        lower = np.zeros(data.ndim)
    if upper == None:
        upper = data.shape
    idx = [lower, upper]
    idx_f = np.fliplr(idx)
    values = wcs.wcs_pix2world(idx_f, 0)
    values = np.fliplr(values)
    return (_unitize(values[0], wcs), _unitize(values[1], wcs))


@support_nddata
def center(data, wcs=None):
    """
        Get center of the data

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.

        Returns
        -------
        result: astropy.units.quantity.Quantity
            Center of the data

    """
    # TODO: These can be a decorator
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    val = wcs.wcs.crval[::-1]
    return _unitize(val, wcs)


@support_nddata
def axes_units(data, wcs=None):
    """
        Get units of the axes

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.

        Returns
        -------
        result: (M,N) or (M,N,Z) numpy.ndarray
            Vector with the units of the axes

    """
    # TODO: These can be a decorator (
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    uvec = np.array(wcs.wcs.cunit)[::-1]
    return uvec


@support_nddata
def resolution(data, wcs=None):
    """
        Get the resolution of data

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.

        Returns
        -------
        result: (M,N) or (M,N,Z) numpy.ndarray
            Resolution of the data

    """
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    val = wcs.wcs.cdelt[::-1]
    return _unitize(val, wcs)


@support_nddata
def spectral_velocities(data, wcs=None, fqs=None, fqis=None, restfrq=None):
    """
        Get the spectral velocities from frequencies fqs given a rest
        frequency (by default search for it in the WCS). If fqs is None,
        then frequencies indices (fqis) need to be given.

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.
        fqs : astropy.units.quantity.Quantity
            Array of frequencies with units.
        fqis : list of integers
            Array of frequencies indices
        restfrq : astropy.units.quantity.Quantity
            Rest frequency

        Returns
        -------
        result: astropy.units.quantity.Quantity
            Array of Spectral velocities.


    """
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    if data.ndim != 3:
        log.error("Not spectral axis found.")
        return None
    if restfrq is None:
        restfrq = wcs.wcs.restfrq * u.Hz
    if fqs is None:
        dim = wcs.wcs.spec
        if fqis is None:
            # Semi Hardconded...
            fqis = np.arange(data.shape[data.ndim - dim - 1])
        idx = np.zeros((fqis.size, data.ndim))
        idx[:, dim] = fqis
        vals = wcs.all_pix2world(idx, 0)
        fqs = vals[:, dim] * u.Hz
    eq = u.doppler_radio(restfrq)
    return fqs.to(u.km / u.s, equivalencies=eq)


@support_nddata
def features(data, wcs=None, region=None):
    """
        Creates an array with WCS axes in features format

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.
        region :(lower : (M,N) or (M,N,Z), upper : (M,N) or (M,N,Z))
            Start and End index in data (int tuples)

        Returns
        -------
        result: astropy.table.Table
            Table with WCS information of a section from the data.

    """
    if region  is None:
        lower = None
        upper = None
    else:
        (lower, upper) = region
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    ii = core.index_features(data, lower, upper)
    f = wcs.wcs_pix2world(ii.T, 0)
    return _world_table_creator(f, wcs)

def _get_axis(wcs,name):
    return wcs.naxis - wcs.axis_type_names.index(name) - 1

def _is_spectra(data,wcs):
    try:
        freq_axis=_get_axis(wcs,"FREQ")
    except ValueError:
        return False
    for i in range(wcs.naxis):
        if i == freq_axis:
            continue
        if data.shape[i]!= 1:
            return False
    return True

@support_nddata
# TODO: Consider using "box" structure rather than up and low
def opening(data, center, window, wcs=None):
    """
        Field of view (center +- window) converted to indices

        Parameters
        ----------
        data : (M,N) or (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
            Astronomical data cube.
        center : astropy.units.quantity.Quantity
            Center of the field of view in WCS.
        window : astropy.units.quantity.Quantity
            Window for the field in WCS.
        wcs : astropy.wcs.wcs.WCS
            World Coordinate System to use.
        Returns
        -------
        result: ((M1,N1,Z1),(M2,N2,Z2)) tuple of tuple of ints
    """
    if wcs is None:
        log.error("A world coordinate system (WCS) is needed")
        return None
    off_low = center - window
    off_low = np.array([x.value for x in off_low])
    off_up = center + window
    off_up = np.array([x.value for x in off_up])
    ld = np.rint(wcs.wcs_world2pix([off_low[::-1]], 0))
    lu = np.rint(wcs.wcs_world2pix([off_up[::-1]], 0))
    lower = np.array([ld, lu]).min(axis=0)
    upper = np.array([ld, lu]).max(axis=0)
    lower = core.fix_limits(data, lower[0][::-1])
    upper = core.fix_limits(data, upper[0][::-1])
    return lower, upper



from astropy import log
import numpy as np
from astropy.wcs import wcs
from astropy.nddata import support_nddata
from smoothy.core.analysis import rms
import matplotlib.pyplot as plt

# TODO: complete the nddata support (i.e. data, meta...)
# TODO: make animation possible again


@support_nddata
def visualize(data,wcs=None,unit=None,contour=False):
    """
    Generic function to visualize data, line-plot for 1D and image for 2D.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        Astronomical image

    wcs : astropy.wcs.WCS
        World Coordinate System from the image (not needed if contained in NDData)

    unit : astropy.unit
        Image units (not needed if contained in NDData)

    contour : numpy.ndarray
        For plotting Contourns
    """
    if data.ndim == 1:
        return visualize_plot(data,wcs,unit)
    elif data.ndim == 2:
        return visualize_image(data,wcs,unit,contour)
    else:
        log.error("Data dimensions must be between 1 and 2")


@support_nddata
def visualize_plot(data,wcs=None,unit=None):
    """
    Plot 1D data for astronomical data.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData
        Astronomical image

    wcs : astropy.wcs.WCS
        World Coordinate System from the image (not needed if contained in NDData)

    unit : astropy.unit
        Image units (not needed if contained in NDData)
    """
    if wcs is None:
        plt.plot(data)
        plt.ylabel(unit)
    else:
        #TODO: Implement x vector, but check why the wcs cannot be onedimensional!
        plt.plot(data)
        plt.ylabel(unit)
        plt.xlabel(wcs.axis_type_names[0])
    #plt.show()


@support_nddata
def visualize_image(data,wcs=None,unit=None,contour=False):
    """
    Plot 2D astronimical data.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        Astronomical image

    wcs : astropy.wcs.WCS
        World Coordinate System from the image (not needed if contained in NDData)

    unit : astropy.unit
        Image units (not needed if contained in NDData)

    contour : numpy.ndarray
        For plotting Contourns
    """
    if wcs is None:
        plt.imshow(data, origin='lower', cmap=plt.cm.gist_heat)
        cb=plt.colorbar()
        cb.ax.set_ylabel(unit)
    else:
        gax=plt.subplot(111,projection=wcs)
        plt.imshow(data, origin='lower', cmap=plt.cm.gist_heat)
        g0=gax.coords[0]
        g1=gax.coords[1]
        g0.set_axislabel(wcs.axis_type_names[0])
        g1.set_axislabel(wcs.axis_type_names[1])
        g0.grid(color='yellow', alpha=0.5, linestyle='solid')
        g1.grid(color='yellow', alpha=0.5, linestyle='solid')
        cb=plt.colorbar()
        cb.ax.set_ylabel(unit)
    if contour:
        arms=rms(data)
        dmax=data.max()
        crs=np.arange(1,dmax/arms)
        plt.contour(data,levels=arms*crs,alpha=0.5)
    plt.show()



from astropy.io import fits
from astropy import log
import numpy as np
import astropy.units as u
from astropy.nddata import support_nddata
from astropy.wcs import wcs
from astropy.table.table import Table

from smoothy.upi.data import Data

@support_nddata
def remove_superfluous(data, wcs=None, mask=None, unit=None,meta=None, restfrq=None):
    if data.shape[0] == 1:
        log.info("Removing superfluous dimension "+wcs.axis_type_names[wcs.naxis - 1]+" (kept in the metadata)")
        cs=wcs.dropaxis(wcs.naxis - 1)
        if mask is not None:
            mk=mask[0]
        dt=data[0]
    else:
        return(Data(data,wcs=wcs,unit=unit,mask=mask,meta=meta))
    newimg=Data(dt,wcs=cs,unit=unit,mask=mk,meta=meta)
    return(remove_superfluous(newimg))


def HDU_to_Data(hdu):
    """
        Create an N-dimensional dataset from an HDU component.

        Parameters
        ----------
        hdu : HDU object
             HDU to transform into an N-dimensional dataset.

        Returns
        -------
        result: astropy.nddata.NDDataRef with data from the HDU object.
   """
    hdu.verify("fix")
    data = hdu.data
    meta = hdu.header
    mask = np.isnan(data)
    # Hack to correct wrong uppercased units generated by CASA
    try:
        bscale = meta['BSCALE']
    except KeyError:
        bscale = 1.0
    try:
        bzero = meta['BZERO']
    except KeyError:
        bzero = 0.0
    try:
        bsu = meta['BUNIT']
        bsu = bsu.lower()
        bsu = bsu.replace("jy", "Jy")
        bunit = u.Unit(bsu, format="fits")
    except KeyError:
        bunit = u.Unit("u.Jy/u.beam")
    rem_list = []
    for i in meta.items():
        if i[0].startswith('PC00'):
            rem_list.append(i[0])
    for e in rem_list:
        meta.remove(e)

    mywcs = wcs.WCS(meta)
    data = data * bscale + bzero
    # Create astropy units
    #if len(data.shape) == 4:
        # Put data in physically-meaninful values, and remove stokes
        # TODO: Stokes is removed by summing (is this correct? maybe is averaging?)
        #log.info("4D data detected: assuming RA-DEC-FREQ-STOKES (like CASA-generated ones), and dropping STOKES")
     #   data = data.sum(axis=0) * bscale + bzero
     #   mask = np.logical_and.reduce(mask,axis=0)
     #   mywcs = mywcs.dropaxis(3)
    #elif len(data.shape) == 3:
    #    log.info("3D data detected: assuming RA-DEC-FREQ")
    #
    #elif len(data.shape) == 2:
    #    log.info("2D data detected: assuming RA-DEC")
    #    data = data * bscale + bzero
    #else:
    #    log.error("Only 3D data allowed (or 4D in case of polarization)")
    #    raise TypeError
    return remove_superfluous(Data(data, uncertainty=None, mask=mask, wcs=mywcs, meta=meta, unit=bunit))

def HDU_to_Table(hdu):
    """
        Create a data table from a HDU component.

        Parameters
        ----------
        hdu : HDU object
            HDU to transform into a data table.

        Returns
        -------
        result: astropy.table.Table with data from the HDU.
    """
    log.warning("FITS Table ---> AstroPy Table not implemented Yet")
    # return atable.ATable(data=hdu.data,meta=hdu.header)


def Table_to_HDU(tab):
    """
        Create a HDU object from a data table.

        Parameters
        ----------
        tab : astropy.table.Table
            Table to transform into a HDU object.

        Returns
        -------
        result: HDU object with data from the data table.
    """
    # if tab.data.masked:???
    #    dtmp = [col.filled(None) for col in six.itervalues(self.columns)]
    hdu = fits.BinTableHDU.from_columns(np.array(tab))
    if tab.meta is not None:
        for k, v in tab.meta.items():
            hdu.header[k] = v
    return hdu


def Data_to_HDU(cube, primary=False):
    """
        Create a HDU object from an N-dimensional dataset.

        Parameters
        ----------
        cube : numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDDataRef
            Astronomical data cube.
        primary : bool
            Whether to pick the primary or an image HDU.

        Returns
        -------
        result: HDU object with data from the data cube.
    """
    header = cube.wcs.to_header()
    if primary == True:
        hdu = fits.PrimaryHDU(cube.data, header=header)
    else:
        hdu = fits.ImageHDU(cube.data, header=header)

    if cube.meta is not None:
        for k, v in cube.meta.items():
            hdu.header[k] = v
    return hdu

def load_fits(filePath, primary = False):
    """
        Loads a FITS file and converts it into an N-Dimensional Dataset.

        Parameters
        ----------
        filepath : path of the FITS file.
        primary : bool
                if True it gets only primmary data-cube.
                
        Returns
        -------
        Primary NDData image or astropy table, and/or:
                N-Dimensional Datasets and Astropy Tables lists
    """
    if primary == True:
        hdulist = fits.open(filePath, lazy_load_hdus=True)
        log.info('Processing PrimaryHDU Object 0')
        hduobject = hdulist[0]
        if hduobject is None:
            log.error('FITS PrimaryHDU is None')
            raise ValueError('FITS PrimaryHDU is None')
        NDData = HDU_to_Data(hduobject)
        hdulist.close()
        return NDData
    else:
        NDData_list = [None]
        Table_list = []
        hdulist = fits.open(filePath)
        for counter, hdu in enumerate(hdulist):
            if isinstance(hdu, fits.PrimaryHDU) or isinstance(hdu, fits.ImageHDU):
                log.info("Processing HDU " + str(counter) + " (Image)")
                try:
                    ndd = HDU_to_Data(hdu)
                    if isinstance(hdu, fits.PrimaryHDU):
                        NDData_list[0] = ndd
                        primary = ndd
                    else:
                        NDData_list[0] = None
                except TypeError:
                    log.info(str(counter) + " (Image) wasn't an Image")
            if isinstance(hdu, fits.BinTableHDU):
                table = HDU_to_Table(hdu)
                Table_list.append(table)
        if NDData_list[0] is None:
            if len(NDData_list) == 0:
                primary = Table_list[0]
            else:
                del NDData_list[0]
                primary = NDData_list[0]

        return primary, NDData_list, Table_list



from astropy import nddata as ndd

from smoothy.upi import axes


class Data(ndd.NDDataRef):
    """
    A generic represenation of astronomical n-dimensional data array. Extends NDData.

    """

    def axes_names(self):
        """
            Get the axes's names.
            
            Returns
            -------
            result: numpy.ndarray
                Numpy ndarray with the axes's names from the WCS.
        """

        return axes.axes_names(self)

    def extent(self, region=None):
        """
            Get the axes extent.

            Parameters
            ----------
            region :(lower : (M,N) or (M,N,Z), upper : (M,N) or (M,N,Z))
                Start and End index in data (int tuples)

            Returns
            -------
            result: (M, N) tuple of astropy.units.quantity.Quantity
                Axes extent
        """
        return axes.extent(self, region)

    def center(self):
        """
            Get center of the data

            Returns
            -------
            result: astropy.units.quantity.Quantity
                Center of the data
        """
        return axes.center(self)

    def axes_units(self):
        """
            Get units of the axes

            Returns
            -------
            result: (M,N) or (M,N,Z) numpy.ndarray
                Vector with the units of the axes
        """
        return axes.axes_units(self)

    def resolution(self):
        """
            Get the resolution of data

            Returns
            -------
            result: (M,N) or (M,N,Z) numpy.ndarray
                Resolution of the data
        """
        return axes.resolution(self)

    def spectral_velocities(self, fqs=None, fqis=None, restfrq=None):
        """
            Get the spectral velocities from frequencies fqs given a rest
            frequency (by default search for it in the WCS). If fqs is None,
            then frequencies indices (fqis) need to be given.

            Parameters
            ----------
            fqs : astropy.units.quantity.Quantity
                Array of frequencies with units.
            fqis : list of integers
                Array of frequencies indices
            restfrq : astropy.units.quantity.Quantity
                Rest frequency

            Returns
            -------
            result: astropy.units.quantity.Quantity
                Array of Spectral velocities.
        """
        return axes.spectral_velocities(self, fqs=fqs, fqis=fqis, restfrq=restfrq)

    def features(self, region=None):
        """
            Creates an array with WCS axea in features format

            Parameters
            ----------
            region :(lower : (M,N) or (M,N,Z), upper : (M,N) or (M,N,Z))
                Start and End index in data (int tuples)

            Returns
            -------
            result: astropy.table.Table
                Table with WCS information of a section from the data.
        """
        return axes.features(self, region=region)

    def opening(self, center, window):
        """
            Field of view (center +- window) converted to indices

            Parameters
            ----------
            center : astropy.units.quantity.Quantity
                Center of the field of view in WCS.
            window : astropy.units.quantity.Quantity
                Window for the field in WCS.

            Returns
            -------
            result: ((M1,N1,Z1),(M2,N2,Z2)) tuple of tuple of ints
        """
        return axes.opening(self, center=center, window=window)



import numpy as np


def fix_mask(data, mask):
    """

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        Astronomical data cube.
    mask : numpy.ndarray
        Boolean that will be applied.

    Returns
    -------
    result : numpy.ma.MaskedArray
        Masked astronomical data cube.
    """

    ismasked = isinstance(data, np.ma.MaskedArray)
    if ismasked and mask is None:
        return data
    else:
        return np.ma.MaskedArray(data, mask)


def fix_limits(data, vect):
    """
    Fix vect index to be inside data

    Parameters
    ----------
    data : numpy.ndarray or numpy.ma.MaskedArray
        Astronomical data cube.
    vect : tuple, list or numpy.ndarray
        Array with the indexes to be fixed.

    Returns
    -------
    result : numpy.ndarray
        Fixed array of indexes.
    """

    if isinstance(vect, (tuple, list)):
        vect = np.array(vect)
    vect = vect.astype(int)
    low = vect < 0
    up = vect > data.shape
    if vect.any():
        vect[low] = 0
    if vect.any():
        vect[up] = np.array(data.shape)[up]
    return vect


def index_features(data, lower=None, upper=None):
    """ Creates an array with indices in features format """
    msh = index_mesh(data, lower, upper)
    dim = data.ndim
    ii = np.empty((dim, int(msh.size / dim)))
    for i in range(dim):
        ii[dim - i - 1] = msh[i].ravel()
    return ii


def slab(data, lower=None, upper=None):
    """
    Obtain the n-dimensional slab from lower to upper (i.e. slab is a vector of slices)

    Parameters
    ----------
    data : numpy.ndarray
        Atronomical data cube.
    lower : 3-tuple (default=None)
        Lower coordinates for the subcube.

    upper : 3-tuple (default=None)
        Upper coordinates for the subcube.


    Returns
    -------
    result : list
       list of slices using lower and upper coordinates to create a subcube.
    """

    if lower is None:
        lower = np.zeros(data.ndim)
    if upper is None:
        upper = data.shape
    lower = fix_limits(data, lower)
    upper = fix_limits(data, upper)
    m_slab = []
    for i in range(data.ndim):
        m_slab.append(slice(lower[i], upper[i]))
    return m_slab


def standarize(data):
    """
    Standarize astronomical data cubes in the 0-1 range.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDData
        Astronomical data cube.

    Returns
    -------
    result : tuple
        Tuple containing the standarized numpy.ndarray or astropy.nddata.NDData
        cube, the factor scale y_fact and the shift y_min.
    """
    y_min = data.min()
    res = data - y_min
    y_fact = res.sum()
    res = res / y_fact
    return res, y_fact, y_min


def unstandarize(data, a, b):
    """
    Unstandarize the astronomical data cube: :math:`a \cdot data + b`.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDData
        Astronomical data cube.
    a : float
        Scale value.
    b : float
        Shift value.

    Returns
    -------
    result : numpy.ndarray or astropy.nddata.NDData
        Unstandarized astronomical cube.
    """
    return a*data+b


def add(data, flux, lower, upper):
    """
    Adds flux to a sub-cube of an astronomical data cube.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDData
        Astronomical data cube.
    flux : numpy.ndarray
        Flux added to the cube.
    lower : tuple
        Lower bound of the sub-cube to which flux will be added.
    upper : tuple
        Upper bound of the sub-cube to which flux will be added.
    """

    data_slab, flux_slab = matching_slabs(data, flux, lower, upper)
    data[data_slab] += flux[flux_slab]


def matching_slabs(data, flux, lower, upper):
    """
    Obtain the matching subcube inside the lower and upper points.

    Parameters
    ----------
    data : numpy.ndarray
        First data cube

    flux : numpy.ndarray
        Second data cubse

    lower : tuple
        Lower coordinates for the subcube.

    upper : tuple
        Upper coordinates for the subcube.


    Returns
    -------
    The subcube inside the lower and upper points that matches both data cube dimensions.
    """

    data_slab = slab(data, lower, upper)
    flow = np.zeros(flux.ndim)
    fup = np.array(flux.shape)
    for i in range(data.ndim):
        if data_slab[i].start == 0:
            flow[i] = flux.shape[i] - data_slab[i].stop
        if data_slab[i].stop == data.shape[i]:
            fup[i] = data_slab[i].stop - data_slab[i].start
    flux_slab = slab(flux, flow, fup)
    return data_slab, flux_slab



from .utilities import fix_mask, slab
import numpy as np


def rms(data, mask=None):
    """
    Compute the RMS of data. If mask != None, then we use that mask.

    Parameters
    ----------
    data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDDataRef
        Astronomical data cube.

    mask : numpy.ndarray (default = None)

    Returns
    -------
    RMS of the data (float)
    """
    # TODO: check photutils background estimation for using that if possible
    if mask is not None:
        data = fix_mask(data, mask)
    mm = data * data
    rms = np.sqrt(mm.sum() * 1.0 / mm.size)
    return rms


def denoise(data, threshold):
    """
    Performs denoising of data cube, thresholding over the threshold value.

    Parameters
    ----------
    data : numpy.ndarray or astropy.nddata.NDData or or astropy.nddata.NDData
        Astronomical data cube.
    threshold : float
        Threshold value used for denoising.

    Returns
    -------
    result : numpy.ndarray
        Denoised (thresholded) astronomical data cube.
    """

    elms = data > threshold
    newdata = np.zeros(data.shape)
    newdata[elms] = data[elms]
    return newdata


def gaussian_function(mu, P, feat, peak):
    """
    Generates an N-dimensional Gaussian using the feature matrix feat,
    centered at mu, with precision matrix P and with intensity peak.

    Parameters
    ----------
    mu : numpy.ndarray
        Centers of gaussians array.
    P : numpy.ndarray
        Precision matrix.
    feat : numpy.ndarray.
        Features matrix.
    peak : float
        Peak value of the resulting evaluation.

    Returns
    -------
    result: 2D numpy.ndarray
        Returns the gaussian function evaluated at the value on feat.
    """

    cent_feat = np.empty_like(feat)
    for i in range(len(mu)):
        cent_feat[i] = feat[i] - mu[i]
    qform = (P.dot(cent_feat)) * cent_feat
    quad = qform.sum(axis=0)
    res = np.exp(-quad / 2.0)
    res = peak * (res / res.max())
    return res


def integrate(data, mask=None, axis=(0)):
    """
    Sums the slices of a cube of data given an axis.

    Parameters
    ----------
    data : (M,N,Z) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        Astronomical data cube.

    mask : numpy.ndarray (default = None)

    axis : int (default=(0))

    Returns
    -------
     A numpy array with the integration results.

    """
    if mask is not None:
        data = fix_mask(data, mask)
    newdata = np.sum(data, axis=axis)
    mask = np.isnan(newdata)
    return newdata



from setuptools import setup, find_packages


setup(
    name="smoothy",
    version="0.1.0",
    description="Package that provides an easy interface and I/O operations for using astronomical libraries.",
    url="https://github.com/ChileanVirtualObservatory/smoothy",
    author="CSRG",
    author_email='contact@lirae.cl',
    classifiers=[
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering :: Astronomy',

        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6'
        ],
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    install_requires=['numpy>=1.11.2', 'astropy>=1.2', 'matplotlib>=1.5']
)



from astropy.nddata import support_nddata, NDDataRef
from smoothy.core import rms, fix_mask, standarize, unstandarize, add, denoise, gaussian_function
import numpy as np

from smoothy.upi.axes import opening, features, axes_units


@support_nddata
def noise_level(data,mask=None,unit=None):
    """
        Compute the RMS of data.

        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        mask : numpy.ndarray
            mask for the data
        unit : astropy.units.Unit
            Astropy Unit (http://docs.astropy.org/en/stable/units/)

        Returns
        -------
        rms : float
            RMS of data
    """

    # TODO: check photutils background estimation for using that if possible
    if unit is None:
        return rms(data,mask)
    else:
        return rms(data,mask)*unit


@support_nddata
def standarize(data, wcs=None, unit=None, mask=None, meta=None):
    """
        Standarize data:

        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        wcs : World Coordinate System data (http://docs.astropy.org/en/stable/wcs/)
        mask : numpy.ndarray
            mask for the data
        unit : astropy.units.Unit
            Astropy Unit (http://docs.astropy.org/en/stable/units/)
        meta : FITS metadata

        Returns
        -------
        Standarized data where data = a * res + b
    """
    if mask is not None:
        data = fix_mask(data, mask)
    (res, a, b) = standarize(data)
    res = NDDataRef(res, uncertainty=None, mask=mask, wcs=wcs, meta=meta, unit=unit)
    return res, a, b


@support_nddata
def unstandarize(data, a, b, wcs=None, unit=None, mask=None, meta=None):
    """
        Unstandarize data: res = a * data + b


        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        a : float
            slope of straight
        b : float
            Intercept of straight
        wcs : World Coordinate System data (http://docs.astropy.org/en/stable/wcs/)
        mask : numpy.ndarray
            mask for the data
        unit : astropy.units.Unit
            Astropy Unit (http://docs.astropy.org/en/stable/units/)
        meta : FITS metadata

        Returns
        -------
        NDDataRef: Unstandarized data: res = a * data + b
    """
    if mask is not None:
        data = fix_mask(data, mask)
    res = unstandarize(data, a, b)
    return NDDataRef(res, uncertainty=None, mask=mask, wcs=wcs, meta=meta, unit=unit)


@support_nddata
def add(data, flux, lower=None, upper=None,wcs=None,unit=None,meta=None,mask=None):
    """
        Create a new data with the new flux added.

        Lower and upper are bounds for data. This operation is border-safe and creates a new object at each call.

        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        flux : float
            Flux of data
        lower : numpy.ndarray
        upper : numpy.ndarray
            Bounds for data
        wcs : World Coordinate System data (http://docs.astropy.org/en/stable/wcs/)
        mask : numpy.ndarray
            mask for the data
        unit : astropy.units.Unit
            Astropy Unit (http://docs.astropy.org/en/stable/units/)
        meta : FITS metadata

        Returns
        -------
        NDDataRef: structure with new flux added

    """

    # Please use the OO version data.add(flux) for modifying the data itself.
    res = data.copy()
    add(res, flux, lower, upper)
    return NDDataRef(res, uncertainty=None, mask=mask, wcs=wcs, meta=None, unit=unit)


@support_nddata
def denoise(data, wcs=None, mask=None, unit=None, threshold=0.0):
    """
        Simple denoising given a threshold (creates a new object)

        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        wcs : World Coordinate System data (http://docs.astropy.org/en/stable/wcs/)
        mask : numpy.ndarray
            mask for the data
        unit : astropy.units.Unit
            Astropy Unit (http://docs.astropy.org/en/stable/units/)
        threshold : float

        Returns
        -------
        NDDataRef: Data denoised

    """
    newdata = denoise(data, threshold.value)
    return NDDataRef(newdata, uncertainty=None, mask=mask, wcs=wcs, meta=None, unit=unit)


@support_nddata
def world_gaussian(data, mu, P, peak, cutoff, wcs=None):
    """
        Creates a gaussian flux at mu position (WCS), with P shape, with a maximum value equal to peak,
        and with compact support up to the cutoff contour

        Parameters
        ----------
        data : (M,N) numpy.ndarray or astropy.nddata.NDData or astropy.nddata.NDDataRef
        mu : float
        P : tuple
            Shape of result
        peak : float
            maximum value

        cutoff :

        wcs : World Coordinate System data (http://docs.astropy.org/en/stable/wcs/)

        Returns
        -------

        Tuple of gaussian flux and borders

    """
    Sigma = np.linalg.inv(P)
    window = np.sqrt(2 * np.log(peak / cutoff) * np.diag(Sigma)) * axes_units(data, wcs=wcs)
    lower, upper = opening(data, mu, window, wcs=wcs)
    if np.any(np.array(upper - lower) <= 0):
        return None, lower, upper
    feat = features(data, wcs=wcs, lower=lower, upper=upper)
    feat = np.array(feat.columns.values())
    mu = np.array([x.value for x in mu])
    res = gaussian_function(mu, P, feat, peak)
    # TODO Not generic
    res = res.reshape(upper[0] - lower[0], upper[1] - lower[1], upper[2] - lower[2])
    return res, lower, upper
